@@ -18,6 +18,7 @@
 #include <errno.h>
 #include <inttypes.h>
 #include <unistd.h>
+#include <linux/dcbnl.h>
 #include <linux/genetlink.h>
 #include <linux/mlxdevm_netlink.h>
 #include <libmnl/libmnl.h>
@@ -153,6 +154,8 @@ static struct ifname_map *ifname_map_alloc(const char *bus_name,
 #define MLXDEVM_OPT_PORT_FN_RATE_PARENT_NODE_NAME BIT(17)
 #define MLXDEVM_OPT_PORT_FUNCTION_TRUST_STATE   BIT(18)
 #define MLXDEVM_OPT_PORT_FN_CAP_ESWITCH		BIT(19)
+#define MLXDEVM_OPT_PORT_FN_MAX_IO_EQS		BIT(20)
+#define MLXDEVM_OPT_PORT_FN_RATE_TC_BW		BIT(21)
 
 struct mlxdevm_opts {
 	uint64_t present; /* flags of present items */
@@ -173,11 +176,13 @@ struct mlxdevm_opts {
 	uint8_t port_fn_cap_roce;
 	uint32_t port_fn_cap_max_uc_list;
 	uint16_t port_function_rate_type;
+	uint32_t tc_bw[IEEE_8021QAZ_MAX_TCS];
 	char *rate_node_name;
 	uint64_t port_function_rate_tx_share;
 	uint64_t port_function_rate_tx_max;
 	const char *rate_parent_node;
 	uint8_t port_fn_trust;
+	uint32_t port_fn_max_io_eqs;
 };
 
 struct mlxdevm {
@@ -384,6 +389,7 @@ mlxdevm_function_policy[MLXDEVM_PORT_FUNCTION_ATTR_MAX + 1] = {
 	[MLXDEVM_PORT_FN_ATTR_EXT_CAP_UC_LIST] = MNL_TYPE_U32,
 	[MLXDEVM_PORT_FN_ATTR_TRUST_STATE] = MNL_TYPE_U8,
 	[MLXDEVM_PORT_FN_ATTR_EXT_CAP_ESWITCH] = MNL_TYPE_U8,
+	[MLXDEVM_PORT_FN_ATTR_MAX_IO_EQS] = MNL_TYPE_U32,
 };
 
 static int function_attr_cb(const struct nlattr *attr, void *data)
@@ -906,7 +912,8 @@ static const struct mlxdevm_args_metadata mlxdevm_args_required[] = {
 	{MLXDEVM_OPT_PORT_FN_CAP_ROCE,		"Port function's roce capability state is expected."},
 	{MLXDEVM_OPT_PORT_FN_CAP_UC_LIST,	"Port function's max uc list capability value is expected."},
 	{MLXDEVM_OPT_PORT_FUNCTION_TRUST_STATE,	"Port function's trust state expected."},
-	{MLXDEVM_OPT_PORT_FN_CAP_ESWITCH,	"Port function's eswitch capability value is expected."},
+	{MLXDEVM_OPT_PORT_FN_CAP_ESWITCH,       "Port function's eswitch capability value is expected."},
+	{MLXDEVM_OPT_PORT_FN_MAX_IO_EQS,	"Port function's max IO EQs value is expected."},
 };
 
 static int mlxdevm_args_finding_required_validate(uint64_t o_required,
@@ -981,6 +988,85 @@ static int mlxdevm_argv_handle_rate(struct mlxdevm *mlxdevm, char **p_bus_name,
 	return 0;
 }
 
+static int tc_bw_attr_cb(const struct nlattr *attr, void *data)
+{
+	struct nlattr **tb = data;
+
+	if (mnl_attr_type_valid(attr, MLXDEVM_ATTR_RATE_TC_BW_MAX)) {
+		tb[mnl_attr_get_type(attr)] = (struct nlattr *)attr;
+		return MNL_CB_OK;
+	}
+
+	return MNL_CB_ERROR;
+}
+
+static int parse_tc_bw_arg(const char *tc_bw_str, int *tc_index, int *tc_bw)
+{
+	char *index, *value;
+	char *input = NULL;
+	int err;
+
+	input = strdup(tc_bw_str);
+	if (!input)
+		return -ENOMEM;
+
+	err = str_split_by_char(input, &index, &value, ':');
+	if (err) {
+		fprintf(stderr, "Invalid format in token: %s\n", input);
+				free(input);
+		return -EINVAL;
+	}
+
+	*tc_index = atoi(index);
+	*tc_bw = atoi(value);
+
+	free(input);
+
+	return 0;
+}
+
+static int parse_tc_bw_args(struct mlxdevm *mlxdevm, uint32_t *tc_bw)
+{
+	const char *tc_bw_str;
+	int index, bw;
+	int total = 0;
+	int err;
+	int i;
+
+	memset(tc_bw, 0, sizeof(uint32_t) * IEEE_8021QAZ_MAX_TCS);
+
+	for (i = 0; i < IEEE_8021QAZ_MAX_TCS; i++) {
+		err = mlxdevm_argv_str(mlxdevm, &tc_bw_str);
+		if (err)
+			return err;
+
+		err = parse_tc_bw_arg(tc_bw_str, &index, &bw);
+		if (err)
+			return err;
+
+		if (index < 0 || index >= IEEE_8021QAZ_MAX_TCS) {
+			fprintf(stderr, "Invalid index: %d\n", index);
+			return -EINVAL;
+		}
+
+		if ( bw < 0 || bw > 100) {
+			fprintf(stderr, "Invalid bandwidth: %d\n", bw);
+			return -EINVAL;
+		}
+
+		tc_bw[index] = bw;
+		total += bw;
+	}
+
+	if (total && total != 100) {
+		fprintf(stderr,
+		 "Sum of all traffic class bandwidth values must be 100, got: %d\n", total);
+		return -EINVAL;
+	}
+
+	return 0;
+}
+
 static int mlxdevm_argv_parse(struct mlxdevm *mlxdevm, uint64_t o_required,
 			 uint64_t o_optional)
 {
@@ -1179,6 +1265,13 @@ static int mlxdevm_argv_parse(struct mlxdevm *mlxdevm, uint64_t o_required,
 			if (err)
 				return err;
 			o_found |= MLXDEVM_OPT_PORT_FN_RATE_PARENT_NODE_NAME;
+		} else if (mlxdevm_argv_match(mlxdevm, "tc-bw") &&
+			   (o_all & MLXDEVM_OPT_PORT_FN_RATE_TC_BW)) {
+			mlxdevm_arg_inc(mlxdevm);
+			err = parse_tc_bw_args(mlxdevm, opts->tc_bw);
+			if (err)
+				return err;
+			o_found |= MLXDEVM_OPT_PORT_FN_RATE_TC_BW;
 		} else if (mlxdevm_argv_match(mlxdevm, "noparent") &&
 			  (o_all & MLXDEVM_OPT_PORT_FN_RATE_PARENT_NODE_NAME)) {
 			mlxdevm_arg_inc(mlxdevm);
@@ -1198,6 +1291,16 @@ static int mlxdevm_argv_parse(struct mlxdevm *mlxdevm, uint64_t o_required,
 				return err;
 
 			o_found |= MLXDEVM_OPT_PORT_FN_CAP_ESWITCH;
+		} else if (mlxdevm_argv_match(mlxdevm, "max_io_eqs") &&
+			  (o_all & MLXDEVM_OPT_PORT_FN_MAX_IO_EQS)) {
+			uint32_t max_io_eqs;
+
+			mlxdevm_arg_inc(mlxdevm);
+			err = mlxdevm_argv_uint32_t(mlxdevm, &max_io_eqs);
+			if (err)
+				return err;
+			opts->port_fn_max_io_eqs = max_io_eqs;
+			o_found |= MLXDEVM_OPT_PORT_FN_MAX_IO_EQS;
 		} else {
 			pr_err("Unknown option \"%s\"\n",
 			       mlxdevm_argv(mlxdevm));
@@ -1228,6 +1331,9 @@ mlxdevm_function_attr_put(struct nlmsghdr *nlh,
 	if (opts->present & MLXDEVM_OPT_PORT_FUNCTION_TRUST_STATE)
 		mnl_attr_put_u8(nlh, MLXDEVM_PORT_FN_ATTR_TRUST_STATE,
 				opts->port_fn_trust);
+	if (opts->present & MLXDEVM_OPT_PORT_FN_MAX_IO_EQS)
+		mnl_attr_put_u32(nlh, MLXDEVM_PORT_FN_ATTR_MAX_IO_EQS,
+				 opts->port_fn_max_io_eqs);
 	mnl_attr_nest_end(nlh, nest);
 }
 
@@ -1276,7 +1382,8 @@ static void mlxdevm_opts_put(struct nlmsghdr *nlh, struct mlxdevm *mlxdevm)
 
 	if (opts->present & (MLXDEVM_OPT_PORT_FUNCTION_HW_ADDR |
 			     MLXDEVM_OPT_PORT_FUNCTION_STATE |
-			     MLXDEVM_OPT_PORT_FUNCTION_TRUST_STATE))
+			     MLXDEVM_OPT_PORT_FUNCTION_TRUST_STATE |
+			     MLXDEVM_OPT_PORT_FN_MAX_IO_EQS))
 		mlxdevm_function_attr_put(nlh, opts);
 	if (opts->present & MLXDEVM_OPT_PORT_FLAVOUR)
 		mnl_attr_put_u16(nlh, MLXDEVM_ATTR_PORT_FLAVOUR,
@@ -1312,6 +1419,18 @@ static void mlxdevm_opts_put(struct nlmsghdr *nlh, struct mlxdevm *mlxdevm)
 	if (opts->present & MLXDEVM_OPT_PORT_FN_RATE_PARENT_NODE_NAME)
 		mnl_attr_put_strz(nlh, MLXDEVM_ATTR_EXT_RATE_PARENT_NODE_NAME,
 				  opts->rate_parent_node);
+	if (opts->present & MLXDEVM_OPT_PORT_FN_RATE_TC_BW) {
+		struct nlattr *nla_tc_bw = mnl_attr_nest_start(nlh, MLXDEVM_ATTR_RATE_TC_BW);
+		int i;
+
+		for (i = 0; i < IEEE_8021QAZ_MAX_TCS; i++) {
+			if (opts->tc_bw[i]) {
+				mnl_attr_put_u32(nlh, i + MLXDEVM_ATTR_RATE_TC_0_BW,
+						 opts->tc_bw[i]);
+			}
+		}
+		mnl_attr_nest_end(nlh, nla_tc_bw);
+	}
 }
 
 static int mlxdevm_argv_parse_put(struct nlmsghdr *nlh,
@@ -1382,6 +1501,26 @@ static void pr_out_rate(struct mlxdevm *mlxdevm, struct nlattr **tb)
 		print_uint(PRINT_ANY, "tx_share", " tx_share %u", tx_share);
 	}
 
+	if (tb[MLXDEVM_ATTR_RATE_TC_BW]) {
+		struct nlattr *tb_tc_bw[MLXDEVM_ATTR_RATE_TC_BW_MAX + 1] = {};
+		int i;
+
+		mnl_attr_parse_nested(tb[MLXDEVM_ATTR_RATE_TC_BW], tc_bw_attr_cb, tb_tc_bw);
+		for (i = 0; i < IEEE_8021QAZ_MAX_TCS; i++) {
+			int j = i + MLXDEVM_ATTR_RATE_TC_0_BW;
+
+			if (tb_tc_bw[j]) {
+				uint32_t  tc_bw = mnl_attr_get_u32(tb_tc_bw[j]);
+				char buf[32];
+
+				if (tc_bw) {
+					snprintf(buf, sizeof(buf), " tc_%d bw %u", i, tc_bw);
+					print_string(PRINT_ANY, NULL, "%s", buf);
+				}
+			}
+		}
+	}
+
 	if (rate_type == MLXDEVM_RATE_EXT_TYPE_LEAF)
 		if (tb[MLXDEVM_ATTR_EXT_RATE_PARENT_NODE_NAME])
 			print_string(PRINT_ANY, "parent", " parent %s", mnl_attr_get_str(tb[MLXDEVM_ATTR_EXT_RATE_PARENT_NODE_NAME]));
@@ -1393,8 +1532,8 @@ static void cmd_port_function_rate_help(void)
 {
 	pr_err("Usage: mlxdevm port function rate add DEV/GROUP_NAME\n");
 	pr_err("Usage: mlxdevm port function rate del DEV/GROUP_NAME\n");
-	pr_err("Usage: mlxdevm port function rate set DEV/GROUP_NAME [ tx_share BW_in_Mbps ] [ tx_max BW_in_Mbps ]\n");
-	pr_err("Usage: mlxdevm port function rate set DEV/PORT_INDEX [ tx_share BW_in_Mbps ] [ tx_max BW_in_Mbps ] [parent group_name]\n");
+	pr_err("Usage: mlxdevm port function rate set DEV/GROUP_NAME [ tx_share BW_in_Mbps ] [ tx_max BW_in_Mbps ] [ tc-bw INDEX:N ... INDEX:N ]\n");
+	pr_err("Usage: mlxdevm port function rate set DEV/PORT_INDEX [ tx_share BW_in_Mbps ] [ tx_max BW_in_Mbps ] [ tc-bw INDEX:N ... INDEX:N ] [parent group_name]\n");
 	pr_err("Usage: mlxdevm port function rate set DEV/PORT_INDEX noparent\n");
 	pr_err("Usage: mlxdevm port function rate show\n");
 	pr_err("Usage: mlxdevm port function rate show DEV/GROUP_NAME\n");
@@ -1461,6 +1600,7 @@ static int cmd_port_function_rate_set(struct mlxdevm *mlxdevm)
 				     MLXDEVM_OPT_HANDLEP | MLXDEVM_OPT_PORT_FN_RATE_NODE_NAME,
 				     MLXDEVM_OPT_PORT_FN_RATE_TX_SHARE |
 				     MLXDEVM_OPT_PORT_FN_RATE_TX_MAX |
+				     MLXDEVM_OPT_PORT_FN_RATE_TC_BW |
 				     MLXDEVM_OPT_PORT_FN_RATE_PARENT_NODE_NAME);
 
 	if (err) {
@@ -2594,6 +2734,13 @@ static void pr_out_port_function(struct mlxdevm *mlxdevm,
 		print_string(PRINT_ANY, "eswitch", " eswitch %s", eswitch ?
 			     "true" : "false");
 	}
+	if (tb[MLXDEVM_PORT_FN_ATTR_MAX_IO_EQS]) {
+		uint32_t max_io_eqs;
+
+		max_io_eqs = mnl_attr_get_u32(tb[MLXDEVM_PORT_FN_ATTR_MAX_IO_EQS]);
+
+		print_uint(PRINT_ANY, "max_io_eqs", " max_io_eqs %u", max_io_eqs);
+	}
 
 	if (!mlxdevm->json_output)
 		__pr_out_indent_dec();
@@ -2711,7 +2858,8 @@ static int cmd_port_function_set(struct mlxdevm *mlxdevm)
 	err = mlxdevm_argv_parse_put(nlh, mlxdevm, MLXDEVM_OPT_HANDLEP,
 				      MLXDEVM_OPT_PORT_FUNCTION_HW_ADDR |
 				      MLXDEVM_OPT_PORT_FUNCTION_STATE |
-				      MLXDEVM_OPT_PORT_FUNCTION_TRUST_STATE);
+				      MLXDEVM_OPT_PORT_FUNCTION_TRUST_STATE |
+				      MLXDEVM_OPT_PORT_FN_MAX_IO_EQS);
 	if (err)
 		return err;
 
