@@ -1,5 +1,7 @@
 /*
- * Copyright (c) 2001-2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: NVIDIA CORPORATION & AFFILIATES
+ * Copyright (c) 2001-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-License-Identifier: GPL-2.0-only or BSD-3-Clause
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -39,8 +41,7 @@
 #include <algorithm>
 #include "tcp_base.h"
 
-class tcp_connect : public tcp_base {
-};
+class tcp_connect : public tcp_base {};
 
 /**
  * @test tcp_connect.ti_1
